from . import MAX_SEMITONES_RESOLVE_BELOW

from .scale import DiatonicScale
from .scale import ChromaticScale

from .sequence import Sequence

from functools import wraps

RESOLUTION_METHODS = {}
# http://stackoverflow.com/questions/5910703/howto-get-all-methods-of-a-pyt\
# hon-class-with-given-decorator
# http://stackoverflow.com/questions/5707589/calling-functions-by-array-ind\
# ex-in-python/5707605#5707605


def register_resolution_method(f, *args, **kwargs):
    """Decorator for resolution method functions.

    Functions decorated with this decorator will be registered in the
    `RESOLUTION_METHODS` global dict.
    """

    @wraps(f)
    def decorator(*args, **kwargs):
        return f(*args, **kwargs)

    RESOLUTION_METHODS.update({f.__name__: f})

    return decorator


class Resolution:
    """This class implements methods for different types of question
    resolutions.

    A resolution is an answer to a question. It aims to create a mnemonic on
    how the inverval resvolver to the tonic.
    """

    def __init__(self, method, question):
        """Inits the resolution class.

        Args:
            method (str): The method used in the resolution.
            question (obj): Question object from which to generate the
            resolution sequence.
        """

        self.METHOD = RESOLUTION_METHODS[method]
        self.question = question

    def __call__(self, *args, **kwargs):
        """Calls the resolution method and pass arguments to it.

        Returns a `birdears.Sequence` object with the resolution generated by
        the.method.
        """
        return self.METHOD(question=self.question, *args, **kwargs)


@register_resolution_method
def nearest_tonic(question):
    """Resolution method that resolve the intervals to their nearest tonics.

    Args:
        question (obj): Question object from which to generate the
            resolution sequence. (this is provided by the `Prequestion` class
            when it is `__call__`ed)
    """

    global DIATONIC_MODES, MAX_SEMITONES_RESOLVE_BELOW

    mode = question.mode
    tonic = question.tonic
    if not hasattr(question, 'question_phrase_intervals'):
        intervals = question.interval
    else:
        intervals = question.question_phrase_intervals
    duration = question.durations['resol']['duration']
    delay = question.durations['resol']['delay']
    pos_delay = question.durations['resol']['pos_delay']
    harmonic = question.is_harmonic
    descending = question.is_descending

    sequence_list = []

    if type(intervals) is not list:
        intervals = [intervals]

    for interval in intervals:
        resolution_pitch = []
        scale_pitch = DiatonicScale(tonic=tonic, mode=mode,
                                    octave=interval.interval_octave,
                                    descending=descending)

        if interval.chromatic_offset <= MAX_SEMITONES_RESOLVE_BELOW:
            begin_to_diatonic = slice(None, interval.diatonic_index + 1)
            resolution_pitch.extend(scale_pitch.scale[begin_to_diatonic])
            if interval.is_chromatic:
                resolution_pitch.append(interval.note_and_octave)
            resolution_pitch.reverse()
        else:
            diatonic_to_end = slice(interval.diatonic_index, None)
            if interval.is_chromatic:
                resolution_pitch.append(interval.note_and_octave)
            resolution_pitch.extend(scale_pitch.scale[diatonic_to_end])

        # unisson and octave
        if interval.semitones == 0:
            resolution_pitch.append(scale_pitch.scale[0])

        elif interval.semitones % 12 == 0:
            # FIXME: multipe octaves
            resolution_pitch.append("{}{}".format(tonic,
                                    interval.tonic_octave))

        if harmonic:
            seq = [[interval.tonic_note_and_octave, x]
                   for x in resolution_pitch]
        else:
            seq = resolution_pitch

        seq_list = list()
        last_el_idx = len(seq) - 1
        for cur_el_idx, cur_el in enumerate(seq):
            cur_delay = delay if last_el_idx != cur_el_idx else pos_delay
            cur_tuple = (cur_el, duration, cur_delay)
            seq_list.append(cur_tuple)

        cur_delay = delay if last_el_idx != cur_el_idx else pos_delay
        cur_tuple = (cur_el, duration, cur_delay)

        sequence_list.extend(seq_list)

    return Sequence(elements=sequence_list, delay=delay, pos_delay=pos_delay)


@register_resolution_method
# FIXME : it should both play preq and question
def repeat_only(question):
    """Resolution method that only repeats the sequence elements with given
    durations.

    Args:
        question (obj): Question object from which to generate the
            resolution sequence. (this is provided by the `Prequestion` class
            when it is `__call__`ed)
    """
    elements = question.question.elements
    duration = question.durations['resol']['duration']
    delay = question.durations['resol']['delay']
    pos_delay = question.durations['resol']['pos_delay']

    sequence_list = Sequence(elements, duration=duration, delay=delay,
                             pos_delay=pos_delay)

    return sequence_list



import subprocess
import time

from random import randrange
from random import choice

from . import KEYBOARD_INDICES
from . import CIRCLE_OF_FIFTHS
from . import DIATONIC_MODES
from . import CHROMATIC_TYPE
from . import INTERVAL_INDEX

from .scale import DiatonicScale
from .scale import ChromaticScale

QUESTION_CLASSES = {}


def register_question_class(f, *args, **kwargs):
    """Decorator for question classes.

    Classes decorated with this decorator will be registered in the
    `QUESTION_CLASSES` global.
    """

    @wraps(f)
    def decorator(*args, **kwargs):
        return f(*args, **kwargs)

    QUESTION_CLASSES.update({f.__name__: f})

    return decorator


class QuestionBase:
    """
    Base Class to be subclassed for Question classes.

    This class implements attributes and routines to be used in Question
    subclasses.
    """

    def __init__(self, mode='major', tonic=None, octave=None, descending=None,
                 chromatic=None, n_octaves=None, valid_intervals=None,
                 user_durations=None, prequestion_method=None,
                 resolution_method=None, default_durations=None,
                 *args, **kwargs):
        """Inits the class.

        Args:
            mode (str): A string represnting the mode of the question.
                Eg., 'major' or 'minor'
            tonic (str): A string representing the tonic of the
                question, eg.: 'C'; if omitted, it will be selected
                randomly.
            octave (int): A scienfic octave notation, for example,
                4 for 'C4'; if not present, it will be randomly chosen.
            descending (bool): Is the question direction in descending,
                ie., intervals have lower pitch than the tonic.
            chromatic (bool): If the question can have (True) or not
                (False) chromatic intervals, ie., intervals not in the
                diatonic scale of tonic/mode.
            n_octaves (int): Maximum numbr of octaves of the question.
            valid_intervals (list): A list with intervals (int) valid for
                random choice, 1 is 1st, 2 is second etc. Eg. [1, 4, 5] to
                allow only tonics, fourths and fifths.
            user_durations (str): A string with 9 comma-separated `int` or
                `float`s to set the default duration for the notes played. The
                values are respectively for: pre-question duration (1st),
                pre-question delay (2nd), and pre-question pos-delay (3rd);
                question duration (4th), question delay (5th), and question
                pos-delay (6th); resolution duration (7th), resolution
                delay (8th), and resolution pos-delay (9th).
                duration is the duration in of the note in seconds; delay is
                the time to wait before playing the next note, and pos_delay is
                the time to wait after all the notes of the respective sequence
                have been played. If any of the user durations is `n`, the
                default duration for the type of question will be used instead.
                Example::
                    "2,0.5,1,2,n,0,2.5,n,1"
            prequestion_method (str): Method of playing a cadence or the
                exercise tonic before the question so to affirm the question
                musical tonic key to the ear. Valid ones are registered in the
                `birdears.prequestion.PREQUESION_METHODS` global dict.
            resolution_method (str): Method of playing the resolution of an
                exercise Valid ones are registered in the
                `birdears.resolution.RESOLUTION_METHODS` global dict.
            user_durations (dict): Dictionary with the default durations for
                each type of sequence. This is provided by the subclasses.
        """

        global KEYBOARD_INDICES, CIRCLE_OF_FIFTHS

        self.mode = mode

        self.is_descending = descending
        self.is_chromatic = chromatic

        # self.octave = octave if octave else randrange(3, 5)
        if not octave:
            octave = randrange(3, 5)
        elif type(octave) == list:
            octave = choice(octave)
        elif type(octave) == tuple and len(octave) == 2:
            octave = randrange(*octave)

        self.octave = octave

        if not n_octaves:
            self.n_octaves = 1
        else:
            self.n_octaves = n_octaves

        # this should go to questionbase
        if type(valid_intervals) == str:
            valid_intervals = valid_intervals.split(',')
            valid_intervals = [int(x, 10) for x in valid_intervals]

        self.valid_intervals = valid_intervals

        direction = 'ascending' if not self.is_descending else 'descending'
        # FIXME: maybe this should go to __main__
        self.keyboard_index = \
            KEYBOARD_INDICES['chromatic'][direction][self.mode]

        if not tonic:
            tonic = choice(CIRCLE_OF_FIFTHS[randrange(2)])
        elif type(tonic) == list:
            tonic = choice(tonic)

        self.tonic = tonic

        diatonic_scale = DiatonicScale(tonic=tonic, mode=mode, octave=None,
                                       descending=descending,
                                       n_octaves=n_octaves)

        chromatic_scale = ChromaticScale(tonic=tonic, octave=None,
                                         descending=descending,
                                         n_octaves=n_octaves)

        diatonic_scale_pitch = DiatonicScale(tonic=tonic, mode=mode,
                                             octave=self.octave,
                                             descending=descending,
                                             n_octaves=n_octaves)

        chromatic_scale_pitch = ChromaticScale(tonic=tonic, octave=self.octave,
                                               descending=descending,
                                               n_octaves=n_octaves)

        scales = dict({
            'diatonic': diatonic_scale,
            'chromatic': chromatic_scale,
            'diatonic_pitch': diatonic_scale_pitch,
            'chromatic_pitch': chromatic_scale_pitch,
        })
        self.scales = scales

        self.concrete_tonic = scales['diatonic_pitch'].scale[0]
        self.scale_size = len(scales['diatonic'].scale)

        self.durations = default_durations
        if user_durations:
            ud_index = {
                0: ('preq', 'duration'),
                1: ('preq', 'delay'),
                2: ('preq', 'pos_delay'),
                3: ('quest', 'duration'),
                4: ('quest', 'delay'),
                5: ('quest', 'pos_delay'),
                6: ('resol', 'duration'),
                7: ('resol', 'delay'),
                8: ('resol', 'pos_delay'),
            }
            ud = user_durations.split(',')
            if len(ud) == len(ud_index):
                for idx, v in ud_index.items():
                    cur_duration = ud[idx].strip()
                    if cur_duration != 'n':
                        self.durations[v[0]][v[1]] = float(cur_duration)

        self.prequestion_method = prequestion_method
        self.resolution_method = resolution_method

    def get_valid_semitones(self):
        """Returns a list with valid semitones for question.
        """

        global DIATONIC_MODES, CHROMATIC_TYPE, MAX_SEMITONES_RESOLVE_BELOW
        global INTERVALS

        diatonic_mode = list(DIATONIC_MODES[self.mode])
        chromatic_network = list(CHROMATIC_TYPE)

        step_network = diatonic_mode

        # FIXME: please refactore this with method signature n_octaves=1:
        if self.n_octaves:
            for i in range(1, self.n_octaves):
                step_network.extend([semitones + 12 * i for semitones in
                                     diatonic_mode[1:]])
                chromatic_network.extend([semitones + 12 * i for semitones in
                                          CHROMATIC_TYPE[1:]])

        if not self.is_chromatic:
            if not self.valid_intervals:
                # semitones = choice(step_network)
                valid_network = step_network
            else:
                valid_network = []
                for item in self.valid_intervals:
                    valid_network.extend(INTERVAL_INDEX[item])
                for i in range(1, self.n_octaves):
                    valid_network.extend([semitones + 12*i for semitones in
                                          valid_network[1:]])
                valid_network = [x for x in valid_network if x in step_network]

                # semitones = choice(valid_network)
        else:
            if not self.valid_intervals:
                # semitones = choice(chromatic_network)
                valid_network = chromatic_network
            else:
                valid_network = []
                for item in self.valid_intervals:
                    valid_network.extend(INTERVAL_INDEX[item])
                for i in range(1, self.n_octaves):
                    valid_network.extend([semitones + 12*i for semitones in
                                          valid_network])

                # semitones = choice(valid_network)

        return valid_network

    def make_question(self):
        """This method should be overwritten by the question subclasses.
        """

        pass

    def make_resolution(self):
        """This method should be overwritten by the question subclasses.
        """

        pass

    def play_question(self):
        """This method should be overwritten by the question subclasses.
        """

        pass

    def check_question(self):
        """This method should be overwritten by the question subclasses.
        """

        pass



from .. import click

from .. import _Getch

from .. import INTERVALS
from .. import DIATONIC_MODES
from .. import CHROMATIC_TYPE

from .. import DEBUG

from os import popen
COLS = int(popen('tput cols', 'r').read())


def center_text(text, sep=True, nl=0):
    """This function returns input text centered according to terminal columns.

    Args:
        text (str): The string to be centered, it can have multiple lines.
        sep (bool): Add line separator after centered text (True) or
            not (False).
        nl (int): How many new lines to add after text.
    """

    linelist = list(text.splitlines())

    # gets the biggest line
    biggest_line_size = 0
    for line in linelist:
        line_lenght = len(line.expandtabs())
        if line_lenght > biggest_line_size:
            biggest_line_size = line_lenght

    columns = COLS
    offset = biggest_line_size / 2
    perfect_center = columns / 2
    padsize = int(perfect_center - offset)
    spacing = ' ' * padsize  # space char

    text = str()
    for line in linelist:
        text += (spacing + line + '\n')

    divider = spacing + ('─' * int(biggest_line_size))  # unicode 0x2500
    text += divider if sep else ''

    text += nl * '\n'

    return text


def print_response(response):
    """Prints the formatted response.

    Args:
        response (dict): A response returned by question's check_question()
    """

    text_kwargs = dict(
        user_resp=response['user_response_str'],
        correct_resp=response['correct_response_str']
    )

    # TODO: make a class for response
    if response['is_correct']:
        response_text = "Correct!"
    else:
        response_text = "Wrong.."

    if 'extra_response_str' in response.keys():
        print(center_text(response['extra_response_str']))

    print(center_text(response_text, nl=2))


def print_instrumental(response):
    """Prints the formatted response for 'instrumental' exercise.

    Args:
        response (dict): A response returned by question's check_question()
    """

    text_kwargs = dict(
        correct_resp=response['correct_response_str']
    )

    response_text = """
{correct_resp}
""".format(**text_kwargs)

    print(center_text(response_text, nl=2))


def print_question(question):
    """Prints the question to the user.

    Args:
        question (obj): A Question class with the question to be printed.
    """

    keyboard = question.keyboard_index

    if not question.is_chromatic:
        scale = list(question.scales['diatonic'].scale)

        mode = list(DIATONIC_MODES[question.mode])
        scale_index = list(mode)
    else:
        scale = list(question.scales['chromatic'].scale)

        mode = list(CHROMATIC_TYPE)
        scale_index = list(mode)

    for o in range(1, question.n_octaves):
        scale_index.extend([x + (12*o) for x in mode[1:]])

    # FIXME: bug with descending n_octaves=2
    if question.is_descending:
        highest = max(scale_index)
        scale_index = [highest - x for x in scale_index]
        scale = reversed(scale)

    intervals = [INTERVALS[i][1] for i in scale_index]
    keys = [keyboard[i] for i in scale_index]

    scale_str = " ".join(map(lambda x: x.ljust(3), scale))
    intervals_str = " ".join(map(lambda x: x.ljust(3), intervals))
    keys_str = " ".join(map(lambda x: x.ljust(3), keys))

    text_kwargs = dict(
        tonic=question.tonic,
        mode=question.mode,
        chroma=question.is_chromatic,
        desc=question.is_descending,
        scale=scale_str,
        intervals=intervals_str,
        keyboard=keys_str,
    )

    question_text = """\

KEY: {tonic} {mode}
(chromatic: {chroma}; descending: {desc})

Intervals {intervals}
Scale     {scale}
Keyboard  {keyboard}

""".format(**text_kwargs)

    print(center_text(question_text, nl=1))


def make_input_str(user_input, keyboard_index):
    """Makes a string representing intervals entered by the user.

    This function is to be used by questions which takes more than one interval
    input as MelodicDictation, and formats the intervals already entered.

    Args:
        user_input (array_type): The list of keyboard keys entered by user.
        keyboard_index (array_type): The keyboard mapping used by question.
    """

    input_str = str()

    user_input_semitones = [keyboard_index.index(s)
                            for s in user_input]

    user_str = "".join([INTERVALS[s][1].center(5)
                       for s in user_input_semitones]).center(COLS)

    input_str = ("\r{}".format(user_str))

    return input_str


def CommandLine(exercise, **kwargs):
    """This function implements the birdears loop for command line.

    Args:
        exercise (str): The question name.
        **kwargs (kwargs): FIXME: The kwargs can contain options for specific
            questions.
    """

    if exercise == 'dictation':
        from ..questions.melodicdictation import MelodicDictationQuestion
        dictate_notes = kwargs['n_notes']
        MYCLASS = MelodicDictationQuestion

    elif exercise == 'instrumental':
        from ..questions.instrumentaldictation \
            import InstrumentalDictationQuestion

        dictate_notes = kwargs['n_notes']
        MYCLASS = InstrumentalDictationQuestion

    elif exercise == 'melodic':
        from ..questions.melodicinterval import MelodicIntervalQuestion
        MYCLASS = MelodicIntervalQuestion
        dictate_notes = 1

    elif exercise == 'harmonic':
        from ..questions.harmonicinterval import HarmonicIntervalQuestion
        MYCLASS = HarmonicIntervalQuestion
        dictate_notes = 1

    getch = _Getch()

    new_question_bit = True

    while True:
        if new_question_bit is True:

            new_question_bit = False

            input_keys = []
            question = MYCLASS(**kwargs)

            print_question(question)
            question.play_question()

        if exercise == 'instrumental':
            response = question.check_question()
            print_instrumental(response)

            new_question_bit = True

            continue

        user_input = getch()

        if user_input in question.keyboard_index and user_input != ' ':  # spc

            input_keys.append(user_input)

            if exercise == 'dictation':
                input_str = make_input_str(input_keys, question.keyboard_index)
                print(input_str, end='')

            if len(input_keys) == dictate_notes:

                response = question.check_question(input_keys)
                print_response(response)

                question.play_resolution()

                new_question_bit = True

        # backspace
        elif user_input == '\x7f':
            if(len(input_keys) > 0) and exercise == 'dictation':
                del(input_keys[-1])
                input_str = make_input_str(input_keys, question.keyboard_index)
                print(input_str, end='')

        # q/Q - quit
        elif user_input in ('q', 'Q'):
            exit(0)

        # r - repeat interval
        elif user_input in ('r', 'R'):
            question.play_question()



from random import choice

from ..questionbase import QuestionBase

from ..interval import DiatonicInterval
from ..interval import ChromaticInterval

from .. import DIATONIC_MODES
from .. import MAX_SEMITONES_RESOLVE_BELOW
from .. import INTERVALS

from ..scale import DiatonicScale
from ..scale import ChromaticScale

from ..sequence import Sequence
from ..resolution import Resolution
from ..prequestion import PreQuestion

from ..interfaces.commandline import COLS
from ..interfaces.commandline import center_text


class InstrumentalDictationQuestion(QuestionBase):
    """Implements an instrumental dictation test.
    """

    def __init__(self, mode='major', wait_time=11, n_repeats=1,
                 max_intervals=3, n_notes=4, tonic=None, octave=None,
                 descending=None, chromatic=None, n_octaves=None,
                 valid_intervals=None, user_durations=None,
                 prequestion_method='progression_i_iv_v_i',
                 resolution_method='repeat_only', *args, **kwargs):
        """Inits the class.

        Args:
            mode (str): A string representing the mode of the question.
                Eg., 'major' or 'minor'.
            wait_time (float): Wait time in seconds for the next question or
                repeat.
            n_repeats (int): Number of times the same dictation will be
                repeated before the end of the exercise.
            max_intervals (int): The maximum number of random intervals the
                question will have.
            n_notes (int): The number of notes the melodic dictation will have.
            tonic (str): A string representing the tonic of the question,
                eg.: 'C'; if omitted, it will be selected randomly.
            octave (int): A scienfic octave notation, for example, 4 for 'C4';
                if not present, it will be randomly chosen.
            descending (bool): Is the question direction in descending, ie.,
                intervals have lower pitch than the tonic.
            chromatic (bool): If the question can have (True) or not (False)
                chromatic intervals, ie., intervals not in the diatonic scale
                of tonic/mode.
            n_octaves (int): Maximum number of octaves of the question.
            valid_intervals (list): A list with intervals (int) valid for
                random choice, 1 is 1st, 2 is second etc. Eg. [1, 4, 5] to
                allow only tonics, fourths and fifths.
            user_durations (str): A string with 9 comma-separated `int` or
                `float`s to set the default duration for the notes played. The
                values are respectively for: pre-question duration (1st),
                pre-question delay (2nd), and pre-question pos-delay (3rd);
                question duration (4th), question delay (5th), and question
                pos-delay (6th); resolution duration (7th), resolution
                delay (8th), and resolution pos-delay (9th).
                duration is the duration in of the note in seconds; delay is
                the time to wait before playing the next note, and pos_delay is
                the time to wait after all the notes of the respective sequence
                have been played. If any of the user durations is `n`, the
                default duration for the type of question will be used instead.
                Example::
                    "2,0.5,1,2,n,0,2.5,n,1"
            prequestion_method (str): Method of playing a cadence or the
                exercise tonic before the question so to affirm the question
                musical tonic key to the ear. Valid ones are registered in the
                `birdears.prequestion.PREQUESION_METHODS` global dict.
            resolution_method (str): Method of playing the resolution of an
                exercise. Valid ones are registered in the
                `birdears.resolution.RESOLUTION_METHODS` global dict.
        """

        default_durations = {
            'preq': {'duration': 2, 'delay': 0.5, 'pos_delay': 1},
            'quest': {'duration': 2, 'delay': 0.5, 'pos_delay': 0},
            'resol': {'duration': 2.5, 'delay': 0.5, 'pos_delay': 1}
        }

        super(InstrumentalDictationQuestion, self).\
            __init__(mode=mode, tonic=tonic, octave=octave,
                     descending=descending, chromatic=chromatic,
                     n_octaves=n_octaves, valid_intervals=valid_intervals,
                     user_durations=user_durations,
                     prequestion_method=prequestion_method,
                     resolution_method=resolution_method,
                     default_durations=default_durations, *args, **kwargs)

        self.is_harmonic = False

        self.wait_time = wait_time
        self.n_repeats = n_repeats

        if not chromatic:
            INTERVAL_CLASS = DiatonicInterval
        else:
            INTERVAL_CLASS = ChromaticInterval

        question_intervals = [INTERVAL_CLASS(mode=mode, tonic=self.tonic,
                              octave=self.octave, n_octaves=self.n_octaves,
                              descending=descending,
                              valid_intervals=self.valid_intervals)
                              for _ in range(max_intervals)]

        self.question_phrase_intervals = [choice(question_intervals)
                                          for _ in range(n_notes-1)]

        self.question_phrase = [0]

        self.question_phrase.extend([interval.semitones
                                     for interval
                                     in self.question_phrase_intervals])

        # self.pre_question = self.make_pre_question(method='none')
        self.pre_question =\
            self.make_pre_question(method=prequestion_method)
        self.question = self.make_question(self.question_phrase)
        self.resolution = self.make_resolution(method=resolution_method)

    def make_pre_question(self, method):
        prequestion = PreQuestion(method=method, question=self)

        return prequestion()

    def make_question(self, phrase_semitones):
        return Sequence([self.scales['chromatic_pitch'].scale[n]
                        for n in phrase_semitones], **self.durations['quest'])

    def make_resolution(self, method):

        resolve = Resolution(method=method, question=self)

        resolution = resolve()

        return resolution

    def play_question(self):

        for r in range(self.n_repeats):
            self.pre_question.play()
            self.question.play()

            for i in range(self.wait_time):
                time_left = str(self.wait_time - i).rjust(3)
                text = '{} seconds remaining...'.format(time_left)
                print(center_text(text, sep=False), end='')
                self.question._wait(1)

    def check_question(self):
        """Checks whether the given answer is correct.

        This currently doesn't applies to instrumental dictation questions.
        """

        global INTERVALS

        intervals_str = "".join([INTERVALS[s][1].center(7)
                                for s in self.question_phrase])
        notes_str = "".join([self.scales['chromatic_pitch'].scale[s].center(7)
                            for s in self.question_phrase])

        correct_response_str = """\
The intervals and notes of this question:

{intervals}
{notes}
""".format(**dict(intervals=intervals_str, notes=notes_str))
        response = {
            'correct_semitones': self.question_phrase,
            'correct_response_str': correct_response_str
        }

        return response



from ..questionbase import QuestionBase

from ..interval import DiatonicInterval
from ..interval import ChromaticInterval

from .. import DIATONIC_MODES
from .. import MAX_SEMITONES_RESOLVE_BELOW
from .. import INTERVALS

from ..scale import DiatonicScale

from ..sequence import Sequence
from ..resolution import Resolution
from ..prequestion import PreQuestion


class HarmonicIntervalQuestion(QuestionBase):
    """Implements a Harmonic Interval test.
    """

    def __init__(self, mode='major', tonic=None, octave=None, descending=None,
                 chromatic=None, n_octaves=None, valid_intervals=None,
                 user_durations=None, prequestion_method='none',
                 resolution_method='nearest_tonic', *args, **kwargs):
        """Inits the class.

        Args:
            mode (str): A string representing the mode of the question.
                Eg., 'major' or 'minor'
            tonic (str): A string representing the tonic of the question,
                eg.: 'C'; if omitted, it will be selected randomly.
            octave (int): A scienfic octave notation, for example, 4 for 'C4';
                if not present, it will be randomly chosen.
            descending (bool): Is the question direction in descending, ie.,
                intervals have lower pitch than the tonic.
            chromatic (bool): If the question can have (True) or not (False)
                chromatic intervals, ie., intervals not in the diatonic scale
                of tonic/mode.
            n_octaves (int): Maximum number of octaves of the question.
            valid_intervals (list): A list with intervals (int) valid for
                random choice, 1 is 1st, 2 is second etc. Eg. [1, 4, 5] to
                allow only tonics, fourths and fifths.
            user_durations (str): A string with 9 comma-separated `int` or
                `float`s to set the default duration for the notes played. The
                values are respectively for: pre-question duration (1st),
                pre-question delay (2nd), and pre-question pos-delay (3rd);
                question duration (4th), question delay (5th), and question
                pos-delay (6th); resolution duration (7th), resolution
                delay (8th), and resolution pos-delay (9th).
                duration is the duration in of the note in seconds; delay is
                the time to wait before playing the next note, and pos_delay is
                the time to wait after all the notes of the respective sequence
                have been played. If any of the user durations is `n`, the
                default duration for the type of question will be used instead.
                Example::
                    "2,0.5,1,2,n,0,2.5,n,1"
            prequestion_method (str): Method of playing a cadence or the
                exercise tonic before the question so to affirm the question
                musical tonic key to the ear. Valid ones are registered in the
                `birdears.prequestion.PREQUESION_METHODS` global dict.
            resolution_method (str): Method of playing the resolution of an
                exercise. Valid ones are registered in the
                `birdears.resolution.RESOLUTION_METHODS` global dict.
        """

        default_durations = {
            'preq': {'duration': 3, 'delay': 0.5, 'pos_delay': 1},
            'quest': {'duration': 3, 'delay': 0.5, 'pos_delay': 0},
            'resol': {'duration': 2.5, 'delay': 0.5, 'pos_delay': 1}
        }

        super(HarmonicIntervalQuestion,
              self).__init__(mode=mode, tonic=tonic, octave=octave,
                             descending=descending, chromatic=chromatic,
                             n_octaves=n_octaves,
                             valid_intervals=valid_intervals,
                             user_durations=user_durations,
                             prequestion_method=prequestion_method,
                             resolution_method=resolution_method,
                             default_durations=default_durations,
                             *args, **kwargs)

        self.is_harmonic = True

        # tonic = self.tonic

        if not chromatic:
            self.interval = \
                DiatonicInterval(mode=mode, tonic=self.tonic,
                                 octave=self.octave,
                                 n_octaves=n_octaves,
                                 descending=descending,
                                 valid_intervals=self.valid_intervals)

        else:
            self.interval = \
                ChromaticInterval(mode=mode, tonic=self.tonic,
                                  octave=self.octave,
                                  n_octaves=n_octaves,
                                  descending=descending,
                                  valid_intervals=self.valid_intervals)

        self.pre_question = self.make_pre_question(method=prequestion_method)
        self.question = self.make_question()
        self.resolution = self.make_resolution(method=resolution_method)

    def make_pre_question(self, method):
        prequestion = PreQuestion(method=method, question=self)

        return prequestion()

    def make_question(self):

        tonic = self.concrete_tonic
        interval = self.interval.note_and_octave
        durations = self.durations

        question = Sequence([[tonic, interval]], **self.durations['quest'])

        return question

    def make_resolution(self, method):

        resolve = Resolution(method=method, question=self)
        resolution = resolve()

        return resolution

    def play_question(self):
        self.pre_question.play()
        self.question.play()

    def play_resolution(self):
        thread = self.resolution.play()
        thread.join()

    def check_question(self, user_input_char):
        """Checks whether the given answer is correct."""

        global INTERVALS

        semitones = self.keyboard_index.index(user_input_char[0])

        tonic = self.scales['chromatic_pitch'].scale[0]

        user_interval = INTERVALS[semitones][2]
        correct_interval = INTERVALS[self.interval.semitones][2]

        user_note = self.scales['chromatic_pitch'].scale[semitones]
        correct_note = self.scales['chromatic_pitch']\
            .scale[self.interval.semitones]

        signal = '✓' if semitones == self.interval.semitones else 'x'  # u2713

        extra_response_str = """\
       “{}” ({}─{})
user {} “{}” ({}─{})
{} semitones
""".format(correct_interval, tonic, correct_note,
           signal, user_interval, tonic, user_note, self.interval.semitones)

        response = dict(
            is_correct=False,
            user_interval=user_interval,
            correct_interval=correct_interval,
            user_response_str=user_interval,
            correct_response_str=correct_interval,
            extra_response_str=extra_response_str,
        )

        if semitones == self.interval.semitones:
            response.update({'is_correct': True})

        else:
            response.update({'is_correct': False})

        return response



import kivy

from kivy.app import App
from kivy.lang import Builder

from kivy.uix.screenmanager import ScreenManager
from kivy.uix.screenmanager import Screen
from kivy.uix.boxlayout import BoxLayout

from kivy.uix.screenmanager import NoTransition

from kivy.uix.button import Button
from kivy.uix.label import Label

from kivy.properties import StringProperty, ObjectProperty

kivy.require('1.10.0')

# Builder.load_file('interfaces/gui/birdears.kv')


class BirdearsScreenManager(ScreenManager):
    def __init__(self, **kwargs):
        super(BirdearsScreenManager, self).__init__(**kwargs)


class InitialScreen(Screen):
    def __init__(self, **kwargs):
        super(InitialScreen, self).__init__(**kwargs)


class ExerciseScreen(Screen):
    def __init__(self, type, **kwargs):
        super(ExerciseScreen, self).__init__(**kwargs)

        self.add_widget(ExerciseWidget(type=type))


class ExerciseWidget(BoxLayout):

    def __init__(self, type, **kwargs):
        super(ExerciseWidget, self).__init__(**kwargs)

        exercise = type
        if exercise == 'melodic':
            from ...questions.melodicinterval import MelodicIntervalQuestion
            self.question = MelodicIntervalQuestion()
        elif exercise == 'harmonic':
            from ...questions.harmonicinterval import HarmonicIntervalQuestion
            self.question = HarmonicIntervalQuestion()
        elif exercise == 'dictation':
            from ...questions.melodicdictation import MelodicDictationQuestion
            self.question = MelodicDictationQuestion()
        elif exercise == 'instrumental':
            from ...questions.instrumentaldictation import\
                InstrumentalDictationQuestion
            self.question = InstrumentalDictationQuestion()

        print(self.question)
        valid_semitones = self.question.get_valid_semitones()

        for semitone in valid_semitones:
            bt = getattr(self, "bt{btid}".format(btid=semitone))

            bt.disabled = False
            bt.keyboard = self.question.keyboard_index[semitone]
            bt.semitone = semitone

            print(bt.keyboard)

        # self.nbprequestion = \
        #     NonblockingSequence(elements=self.question.pre_question.elements,
        #                         duration=self.question.pre_question.duration,
        #                         delay=self.question.pre_question.delay,
        #                         pos_delay=self.question.pre_question.pos_delay)
        #
        # self.nbquestion = \
        #     NonblockingSequence(elements=self.question.question.elements,
        #                         duration=self.question.question.duration,
        #                         delay=self.question.question.delay,
        #                         pos_delay=self.question.question.pos_delay)

        # play_iter = iter(nbquestion)
        # nbquestion.play_next(play_iter)
        self.question.play_question()

    # def play_pre_question(self):
    #     # Clock.schedule_once(lambda dt: self.nbprequestion.play_callback(dt,
    #     #                    self.play_question), 0)
    #     self.question
    #
    # def play_question(self):
    #     Clock.schedule_once(self.nbquestion.play_callback, 0)

    def check_question(self, semitone):

        self.question.play_resolution()
        # self.nbresolution = \
        #     NonblockingSequence(elements=self.question.resolution.elements,
        #                         duration=self.question.resolution.duration,
        #                         delay=self.question.resolution.delay,
        #                         pos_delay=self.question.resolution.pos_delay)
        # Clock.schedule_once(lambda dt: self.nbresolution.play_callback, 0)

        keyboard = self.question.keyboard_index[semitone]
        response = self.question.check_question([keyboard])

        print(response)


class BirdearsApp(App):

    def build(self):

        self.sm = BirdearsScreenManager(transition=NoTransition())
        self.sm.switch_to(InitialScreen(name='initial'))

        return self.sm

    def show_screen(self, screen_name):
        self.sm.current = screen_name

    def show_exercise(self, exercise):

        # if exercise == 'melodic':
        #     from ...questions.melodicinterval import MelodicIntervalQuestion
        #     self.question = MelodicIntervalQuestion()
        # elif exercise == 'harmonic':
        #     from ...questions.harmonicinterval import \
        # HarmonicIntervalQuestion
        #     self.question = HarmonicIntervalQuestion()
        # elif exercise == 'dictation':
        #     from ...questions.melodicdictation import \
        # MelodicDictationQuestion
        #     self.question = MelodicDictationQuestion()
        # elif exercise == 'instrumental':
        #     from ...questions.instrumentaldictation import\
        #         InstrumentalDictationQuestion
        #     self.question = InstrumentalDictationQuestion()

        # self.sm.current = 'exercise'
        # self.sm.current = screen_name
        self.sm.switch_to(ExerciseScreen(name='exercise', type=exercise))



"""
This submodule exports `logger` to log events.

Logging messages which are less severe than `lvl` will be ignored::

    Level       Numeric value
    -----       -------------
    CRITICAL    50
    ERROR       40
    WARNING     30
    INFO        20
    DEBUG       10
    NOTSET      0

    Level       When it’s used
    -----       --------------
    DEBUG       Detailed information, typically of interest only when
                    diagnosing problems.
    INFO        Confirmation that things are working as expected.
    WARNING     An indication that something unexpected happened, or indicative
                    of some problem in the near future (e.g. ‘disk space low’).
                    The software is still working as expected.
    ERROR       Due to a more serious problem, the software has not been able
                    to perform some function.
    CRITICAL	A serious error, indicating that the program itself may be
                    unable to continue running.
"""

import logging
from functools import wraps

from logging.config import dictConfig

logging.getLogger(__name__).addHandler(logging.NullHandler())

log_format = """\
%(levelname)s:%(filename)s,%(lineno)s:%(funcName)s() %(message)s\
"""

date_format = '%Y/%m/%d,%H:%M:%S'

logging_config = dict(
    version=1,
    formatters={
        'f': {
             'format': log_format,
             'datefmt': date_format,
        }
    },
    handlers={
        'h': {
                'class': 'logging.StreamHandler',
                'formatter': 'f',
                'level': logging.DEBUG,
        }
    },
    root={
        'handlers': ['h'],
        'level': logging.WARNING,
    },
)

dictConfig(logging_config)

logger = logging.getLogger()


def log_event(f, *args, **kwargs):
    """
    Decorator. Functions and method decorated with this decorator will have
    their signature logged when birdears is executed with `--debug` mode. Both
    function signature with their call values and their return will be logged.
    """

    @wraps(f)
    def decorator(*args, **kwargs):

        # arguments = str(args) # 0 is self
        # kw_arguments = str(kwargs)
        qname = f.__qualname__

        arguments = ', '.join([repr(arg) for arg in args])  # 0 is self
        kw_arguments = ', '.join(['{}={}'.format(k, repr(v))
                                 for k, v in kwargs.items()])

        if logger.isEnabledFor(logging.INFO):
            logger.info("{qname}() called.".format(qname=qname))

        if logger.isEnabledFor(logging.DEBUG):
            # logger.debug("\t*{}, **{}".format(
            # arguments, kw_arguments).expandtabs())
            logger.debug("{qname}({args}, {kwargs})".
                         format(qname=qname, args=arguments,
                                kwargs=kw_arguments). expandtabs())

            # init returns the very own object
            if f.__name__ != '__init__':

                func_return = f(*args, **kwargs)
                logger.debug("{qname} function returned: '{func_ret}'".
                             format(qname=qname, func_ret=func_return))

                return func_return

        return f(*args, **kwargs)

    # RESOLUTION_METHODS.update({f.__name__: f})

    return decorator



from random import choice

from ..questionbase import QuestionBase

from ..interval import DiatonicInterval
from ..interval import ChromaticInterval

from .. import DIATONIC_MODES
from .. import MAX_SEMITONES_RESOLVE_BELOW
from .. import INTERVALS

from ..scale import DiatonicScale
from ..scale import ChromaticScale

from ..sequence import Sequence
from ..resolution import Resolution
from ..prequestion import PreQuestion


class MelodicDictationQuestion(QuestionBase):
    """Implements a melodic dictation test.
    """

    def __init__(self, mode='major', max_intervals=3, n_notes=4, tonic=None,
                 octave=None, descending=None, chromatic=None,
                 n_octaves=None, valid_intervals=None, user_durations=None,
                 prequestion_method='progression_i_iv_v_i',
                 resolution_method='repeat_only', *args, **kwargs):
        """Inits the class.

        Args:
            mode (str): A string representing the mode of the question.
                Eg., 'major' or 'minor'.
            max_intervals (int): The maximum number of random intervals
                the question will have.
            n_notes (int): The number of notes the melodic dictation will have.
            tonic (str): A string representing the tonic of the question,
                eg.: 'C'; if omitted, it will be selected randomly.
            octave (int): A scienfic octave notation, for example, 4 for 'C4';
                if not present, it will be randomly chosen.
            descending (bool): Is the question direction in descending, ie.,
                intervals have lower pitch than the tonic.
            chromatic (bool): If the question can have (True) or not (False)
                chromatic intervals, ie., intervals not in the diatonic scale
                of tonic/mode.
            n_octaves (int): Maximum number of octaves of the question.
            valid_intervals (list): A list with intervals (int) valid for
                random choice, 1 is 1st, 2 is second etc. Eg. [1, 4, 5] to
                allow only tonics, fourths and fifths.
            user_durations (str): A string with 9 comma-separated `int` or
                `float`s to set the default duration for the notes played. The
                values are respectively for: pre-question duration (1st),
                pre-question delay (2nd), and pre-question pos-delay (3rd);
                question duration (4th), question delay (5th), and question
                pos-delay (6th); resolution duration (7th), resolution
                delay (8th), and resolution pos-delay (9th).
                duration is the duration in of the note in seconds; delay is
                the time to wait before playing the next note, and pos_delay is
                the time to wait after all the notes of the respective sequence
                have been played. If any of the user durations is `n`, the
                default duration for the type of question will be used instead.
                Example::
                    "2,0.5,1,2,n,0,2.5,n,1"
            prequestion_method (str): Method of playing a cadence or the
                exercise tonic before the question so to affirm the question
                musical tonic key to the ear. Valid ones are registered in the
                `birdears.prequestion.PREQUESION_METHODS` global dict.
            resolution_method (str): Method of playing the resolution of an
                exercise. Valid ones are registered in the
                `birdears.resolution.RESOLUTION_METHODS` global dict.
        """

        default_durations = {
            'preq': {'duration': 2, 'delay': 0.5, 'pos_delay': 1},
            'quest': {'duration': 2, 'delay': 0.8, 'pos_delay': 0},
            'resol': {'duration': 2.5, 'delay': 0.5, 'pos_delay': 1}
        }

        super(MelodicDictationQuestion, self).\
            __init__(mode=mode, tonic=tonic, octave=octave,
                     descending=descending, chromatic=chromatic,
                     n_octaves=n_octaves, valid_intervals=valid_intervals,
                     user_durations=user_durations,
                     prequestion_method=prequestion_method,
                     resolution_method=resolution_method,
                     default_durations=default_durations, *args, **kwargs)

        self.is_harmonic = False

        if not chromatic:
            INTERVAL_CLASS = DiatonicInterval
        else:
            INTERVAL_CLASS = ChromaticInterval

        question_intervals = [INTERVAL_CLASS(mode=mode, tonic=self.tonic,
                              octave=self.octave, n_octaves=self.n_octaves,
                              descending=descending,
                              valid_intervals=self.valid_intervals)
                              for _ in range(max_intervals)]

        self.question_phrase_intervals = [choice(question_intervals)
                                          for _ in range(n_notes-1)]

        self.question_phrase = [0]

        self.question_phrase.extend([interval.semitones
                                     for interval
                                     in self.question_phrase_intervals])

        self.pre_question =\
            self.make_pre_question(method=prequestion_method)
        # self.pre_question = self.make_pre_question(method='none')
        self.question = self.make_question(self.question_phrase)
        self.resolution = self.make_resolution(method=resolution_method)

    def make_pre_question(self, method):
        prequestion = PreQuestion(method=method, question=self)

        return prequestion()

    def make_question(self, phrase_semitones):
        return Sequence([self.scales['chromatic_pitch'].scale[n]
                        for n in phrase_semitones], **self.durations['quest'])

    def make_resolution(self, method):

        resolve = Resolution(method=method, question=self)
        resolution = resolve()

        return resolution

    def play_question(self):
        self.pre_question.play()
        self.question.play()

    def play_resolution(self):

        thread = self.resolution.play()
        thread.join()

    def check_question(self, user_input_keys):
        """Checks whether the given answer is correct."""

        global INTERVALS

        STR_OFFSET = 5

        user_input_semitones = [self.keyboard_index.index(s)
                                for s in user_input_keys]

        user_response_str = "".join([INTERVALS[s][1].center(STR_OFFSET)
                                     for s in user_input_semitones])
        correct_response_str = "".join([INTERVALS[s][1].center(STR_OFFSET)
                                        for s in self.question_phrase])

        correct_semitones = list()
        correct_wrong_str = str()

        for i, s in enumerate(self.question_phrase):
            if self.question_phrase[i] == user_input_semitones[i]:
                correct_semitones.append(True)
                correct_wrong_str += "✓".center(STR_OFFSET)  # u2713
            else:
                correct_semitones.append(False)
                correct_wrong_str += "x".center(STR_OFFSET)

        extra_response_str = """\
{}
{}
""".format(correct_wrong_str, correct_response_str)

        response = dict(
            is_correct=False,
            user_input=user_input_keys,
            user_semitones=user_input_semitones,
            question_semitones=self.question_phrase,
            correct_semitones=correct_semitones,
            user_response_str=user_response_str,
            correct_response_str=correct_response_str,
            extra_response_str=extra_response_str,
        )

        if user_input_semitones == self.question_phrase:
            response.update({'is_correct': True})
        else:
            response.update({'is_correct': False})

        return response
